// sessiond - SSL session cache daemon, file comm.cpp
// Copyright (C) 2009 Michal Trojnara <Michal.Trojnara@mirt.net>
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation; either version 2 of the License, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, see <http://www.gnu.org/licenses>.
//
// Linking sessiond statically or dynamically with other modules is making
// a combined work based on sessiond. Thus, the terms and conditions of
// the GNU General Public License cover the whole combination.

#include "data.h"
#include "log.h"
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

static BYTES mem2bytes(const unsigned char *, const unsigned int);
static void bytes2mem(unsigned char *, const BYTES &);

#define CACHE_CMD_NEW     0x00
#define CACHE_CMD_GET     0x01
#define CACHE_CMD_REMOVE  0x02
#define CACHE_RESP_ERR    0x80
#define CACHE_RESP_OK     0x81

#define KEY_LEN 32
#define MAX_VAL_LEN 512
typedef struct {
    u_char version, type;
    u_short timeout;
    u_char key[KEY_LEN];
    u_char val[MAX_VAL_LEN];
} CACHE_PACKET;
static DATA data;

void process_request(const int s, const u_short port, LOG &log) {
    CACHE_PACKET packet;
    struct sockaddr addr;
    socklen_t addrlen;
    ssize_t len=recvfrom(s, &packet, sizeof packet, 0, &addr, &addrlen);
    if(len==-1) {
        log.err(LOG_ERR, "recvfrom");
        sleep(1); // limit the error rate
        return;
    }
    const sockaddr_in *in_addr=(sockaddr_in *)&addr;
    if(!len && in_addr->sin_family==AF_INET &&
            in_addr->sin_port==htons(port) &&
            in_addr->sin_addr.s_addr==htonl(INADDR_LOOPBACK)) {
        log.msg(LOG_INFO, data.stats()); // log statistics
        return;
    }
    if(len<(int)(sizeof packet-MAX_VAL_LEN) || packet.version!=1) {
        log.msg(LOG_ERR, "malformed packet received");
        return;
    }
    BYTES k=mem2bytes(packet.key, KEY_LEN);
    if(packet.type==CACHE_CMD_NEW) {
        BYTES v=mem2bytes(packet.val, len-(sizeof packet-MAX_VAL_LEN));
        data.insert(k, v, ntohs(packet.timeout));
    } else if(packet.type==CACHE_CMD_GET) {
        len=sizeof packet-MAX_VAL_LEN;
        if(data.exists(k)) {
            BYTES v=data.find(k);
            bytes2mem(packet.val, v);
            len+=v.size();
            packet.type=CACHE_RESP_OK;
        } else {
            packet.type=CACHE_RESP_ERR;
        }
        if(sendto(s, &packet, len, 0, &addr, addrlen)==-1)
            log.err(LOG_ERR, "sendto");
    } else if(packet.type==CACHE_CMD_REMOVE) {
        data.erase(k);
    } else {
        log.msg(LOG_ERR, "incorrect packet type");
    }
}

static BYTES mem2bytes(const unsigned char *src, const unsigned int l) {
    BYTES dst(l);
    for(unsigned int i=0; i<l; ++i)
        dst[i]=src[i];
    return dst;
}

static void bytes2mem(unsigned char *dst, const BYTES &src) {
    unsigned int l=src.size();
    for(unsigned int i=0; i<l; ++i)
        dst[i]=src[i];
}

// end of comm.cpp
